package activitypub_client

  1. Overview
  2. Docs

Module type Acti.TSource

This is the signature of the module we get after applying the Make functor on an Object.T: functions to POST activities, using module O to act as a given actor.

module O : Object.T
val accept : accept_fun
val announce : announce_fun
val create : create_fun
val delete : delete_fun
val dislike : dislike_fun
val follow : follow_fun
val like : like_fun
val undo : undo_fun
val update : update_fun