package activitypub_client

  1. Overview
  2. Docs

Class type Actor.oSource

An actor implements AP.Types.actor with additional methods to post activities, create objects or retrieve information.

method as_actor : AP.Types.actor
method create_note : ?add:(Rdf.Graph.graph -> Rdf.Term.term -> unit) -> ?content_map:string AP.Smap.t -> string -> ?in_reply_to:Iri.t list -> Activitypub_client.Acti.audience_post_fun

Create a Note. The add optional argument can be used to insert addition triples in the note object before sending the activity. The term parameter of the add function is the note's node in the graph.

method create_object : AP.Types.object_ -> ?in_reply_to:Iri.t list -> Activitypub_client.Acti.audience_post_fun
method upload_file : (?ct:Ldp.Ct.t -> string -> (Iri.t, AP.E.error) Result.t Lwt.t) option
method followers_list : Iri.t list Lwt.t
method following_list : Iri.t list Lwt.t