Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Acti.TSourceThis is the signature of the module we get after applying the Make functor on an Object.T: functions to POST activities, using module O to act as a given actor.
val accept : accept_funval announce : announce_funval create : create_funval delete : delete_funval dislike : dislike_funval follow : follow_funval like : like_funval undo : undo_funval update : update_fun