Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Actor.oSourceAn actor implements AP.Types.actor with additional methods to post activities, create objects or retrieve information.
inherit AP.Types.actormethod as_actor : AP.Types.actormethod accept : Activitypub_client.Acti.accept_funmethod announce : Activitypub_client.Acti.announce_funmethod create : Activitypub_client.Acti.create_funmethod delete : Activitypub_client.Acti.delete_funmethod dislike : Activitypub_client.Acti.dislike_funmethod follow : Activitypub_client.Acti.follow_funmethod like : Activitypub_client.Acti.like_funmethod undo : Activitypub_client.Acti.undo_funmethod update : Activitypub_client.Acti.update_funmethod create_note : ?add:(Rdf.Graph.graph -> Rdf.Term.term -> unit) ->
?content_map:string AP.Smap.t ->
string ->
?in_reply_to:Iri.t list ->
Activitypub_client.Acti.audience_post_funCreate a Note. The add optional argument can be used to insert addition triples in the note object before sending the activity. The term parameter of the add function is the note's node in the graph.
method create_object : AP.Types.object_ ->
?in_reply_to:Iri.t list ->
Activitypub_client.Acti.audience_post_funmethod create_link : ?height:int ->
?hreflang:string ->
?media_type:Ldp.Ct.mime ->
?name:string ->
?rel:string list ->
?type_:Iri.t ->
?width:int ->
Iri.t ->
?in_reply_to:Iri.t list ->
Activitypub_client.Acti.audience_post_funmethod upload_file : (?ct:Ldp.Ct.t ->
string ->
(Iri.t, AP.E.error) Result.t Lwt.t)
optionmethod followers_list : Iri.t list Lwt.tmethod following_list : Iri.t list Lwt.t